import torch
from emu_base.jump_lindblad_operators import compute_noise_from_lindbladians


dtype = torch.complex128
sigmax = torch.tensor([[0.0, 1.0], [1.0, 0.0]], dtype=dtype)
sigmay = torch.tensor([[0.0, -1.0j], [1.0j, 0.0]], dtype=dtype)
n_op = torch.tensor([[0.0, 0.0], [0.0, 1.0]], dtype=dtype)


class RydbergLindbladian:
    """
    Apply the Lindblad superoperator ℒ to a density matrix 𝜌,  ℒ(𝜌).

    This class implements
    H @𝜌- H @ 𝜌 + i ∑ₖ − 1/2 Aₖ† Aₖ 𝜌 − 1/2 𝜌 Aₖ^† Aₖ + Aₖ 𝜌 Aₖ^†,
    where A_k is a jump operator and H is the Rydberg Hamiltonian.
    The complex -𝑖, will be multiplied in the evolution.

    Attributes:
        nqubits (int): number of qubits in the system.
        omegas (torch.Tensor): amplited frequencies  Ωⱼ for each qubit, divided by 2.
        deltas (torch.Tensor): detunings 𝛿ᵢ for each qubit.
        phis (torch.Tensor): phases 𝜙ᵢ for each qubit.
        interaction_matrix (torch.Tensor): interaction_matrix (torch.Tensor): matrix Uᵢⱼ
            representing pairwise Rydberg interaction strengths between qubits.
        pulser_linblads (list[torch.Tensor]): List of 2x2 local Lindblad (jump)
            operators acting on each qubit.
        device (torch.device): device on which tensors are allocated. cpu or gpu: cuda.
        complex (bool): flag indicating whether any drive phase is nonzero
            (i.e., complex Hamiltonian terms).
        diag (torch.Tensor): precomputed diagonal interaction term for the density matrix evolution.

    Methods:
        apply_local_op_to_density_matrix(density_matrix, local_op, target_qubit):
            Applies a local operator to the density matrix from the left: L @ ρ.

        apply_density_matrix_to_local_op_T(density_matrix, local_op, target_qubit):
            Applies a daggered local operator to the density matrix from the right: ρ @ L†.

        __matmul__(density_matrix):
            Applies the full Lindbladian superoperator to the input density matrix,
            including coherent evolution and all dissipation channels.
    """

    def __init__(
        self,
        omegas: torch.Tensor,
        deltas: torch.Tensor,
        phis: torch.Tensor,
        pulser_linblads: list[torch.Tensor],
        interaction_matrix: torch.Tensor,
        device: torch.device,
    ):
        self.nqubits: int = len(omegas)
        self.omegas: torch.Tensor = omegas / 2.0
        self.deltas: torch.Tensor = deltas
        self.phis: torch.Tensor = phis
        self.interaction_matrix: torch.Tensor = interaction_matrix
        self.pulser_linblads: list[torch.Tensor] = pulser_linblads
        self.device: torch.device = device
        self.complex = self.phis.any()

        self.diag: torch.Tensor = self._create_diagonal()

    def _create_diagonal(self) -> torch.Tensor:
        """
        Return the diagonal elements of the Rydberg Hamiltonian matrix
        concerning only the interaction

            H.diag =  ∑ᵢ﹥ⱼUᵢⱼnᵢnⱼ
        """
        diag = torch.zeros(2**self.nqubits, dtype=dtype, device=self.device)

        for i in range(self.nqubits):
            diag = diag.view(2**i, 2, -1)
            i_fixed = diag[:, 1, :]
            for j in range(i + 1, self.nqubits):
                i_fixed = i_fixed.view(2**i, 2 ** (j - i - 1), 2, -1)
                # replacing i_j_fixed by i_fixed breaks the code :)
                i_j_fixed = i_fixed[:, :, 1, :]
                i_j_fixed += self.interaction_matrix[i, j]
        return diag.view(-1)

    def apply_local_op_to_density_matrix(
        self,
        density_matrix: torch.Tensor,
        local_op: torch.Tensor,
        target_qubit: int,
    ) -> torch.Tensor:
        """
        Calculate a local operator (2x2) L being multiplied by a density matrix ρ
        from the left
        Return L @ ρ
        """

        orignal_shape = density_matrix.shape
        density_matrix = density_matrix.view(2**target_qubit, 2, -1)
        density_matrix = local_op @ density_matrix

        return density_matrix.view(orignal_shape)

    def apply_density_matrix_to_local_op_T(
        self,
        density_matrix: torch.Tensor,
        local_op: torch.Tensor,
        target_qubit: int,
    ) -> torch.Tensor:
        """
        Calculates a density matrix ρ being multiplied by a daggered local (2x2)
          operator L† from the right,

        return:  ρ @L†
        """

        orignal_shape = density_matrix.shape

        density_matrix = density_matrix.view(2 ** (target_qubit + self.nqubits), 2, -1)
        density_matrix = local_op.conj() @ density_matrix

        return density_matrix.view(orignal_shape)

    def __matmul__(self, density_matrix: torch.Tensor) -> torch.Tensor:
        """Apply the i*RydbergLindbladian operator to the density matrix ρ
        in the following way:
        Define and effective Hamiltonian
        Heff = Hρ  -0.5i ∑ₖ Lₖ† Lₖ ρ
        Then, the Lindblad operator applying to ρ is giving by
         ℒ(𝜌) = Heff - Heff^†+i*∑ₖ Lₖ ρ Lₖ†
        """

        # compute -0.5i ∑ₖ Lₖ† Lₖ
        sum_lindblad_local = compute_noise_from_lindbladians(self.pulser_linblads).to(
            self.device
        )

        # apply all local terms:  Ωⱼ σₓ - δⱼ n - 0.5i (∑ₖ Lₖ† Lₖ) to each qubit
        H_den_matrix = torch.zeros_like(density_matrix, dtype=dtype, device=self.device)

        if not self.complex:
            for qubit, (omega, delta) in enumerate(zip(self.omegas, self.deltas)):
                H_q = (
                    omega * sigmax.to(device=self.device)
                    - delta * n_op.to(device=self.device)
                    + sum_lindblad_local
                )
                H_den_matrix += self.apply_local_op_to_density_matrix(
                    density_matrix, H_q, qubit
                )
        else:
            for qubit, (omega, delta, phi) in enumerate(
                zip(self.omegas, self.deltas, self.phis)
            ):
                H_q = (
                    omega
                    * (
                        (
                            torch.cos(phi) * sigmax.to(device=self.device)
                            + torch.sin(phi) * sigmay.to(device=self.device)
                        )
                    )
                    - delta * n_op.to(device=self.device)
                    + sum_lindblad_local
                )
                H_den_matrix += self.apply_local_op_to_density_matrix(
                    density_matrix, H_q, qubit
                )

        # apply the interaction terms  ∑ᵢⱼ Uᵢⱼ nᵢ nⱼ
        H_den_matrix += self.diag.view(-1, 1) * density_matrix

        # Heff - Heff^†=  [H, ρ] - 0.5i ∑ₖ Lₖ† Lₖρ - ρ 0.5i ∑ₖ Lₖ† Lₖρ
        H_den_matrix = H_den_matrix - H_den_matrix.conj().T

        # compute ∑ₖ Lₖ ρ Lₖ†, last part of the Lindblad operator
        L_den_matrix_Ldag = sum(
            self.apply_density_matrix_to_local_op_T(
                self.apply_local_op_to_density_matrix(
                    density_matrix, L.to(self.device), qubit
                ),
                L.to(self.device),
                qubit,
            )
            for qubit in range(self.nqubits)
            for L in self.pulser_linblads
        )

        return H_den_matrix + 1.0j * L_den_matrix_Ldag



import torch

from pulser.backend import (
    CorrelationMatrix,
    EmulationConfig,
    EnergySecondMoment,
    EnergyVariance,
    Occupation,
    Energy,
)

from emu_sv.state_vector import StateVector
from emu_sv.dense_operator import DenseOperator
from emu_sv.hamiltonian import RydbergHamiltonian


def qubit_occupation_sv_impl(
    self: Occupation,
    *,
    config: EmulationConfig,
    state: StateVector,
    hamiltonian: DenseOperator,
) -> torch.Tensor:
    """
    Custom implementation of the occupation ❬ψ|nᵢ|ψ❭ for the state vector solver.
    """
    nqubits = state.n_qudits
    occupation = torch.zeros(nqubits, dtype=torch.float64, device=state.vector.device)
    for i in range(nqubits):
        state_tensor = state.vector.view(2**i, 2, -1)
        # nᵢ is a projector and therefore nᵢ == nᵢnᵢ
        # ❬ψ|nᵢ|ψ❭ == ❬ψ|nᵢnᵢ|ψ❭ == ❬ψ|nᵢ * nᵢ|ψ❭ == ❬ϕ|ϕ❭ == |ϕ|**2
        occupation[i] = torch.linalg.vector_norm(state_tensor[:, 1]) ** 2
    return occupation.cpu()


def correlation_matrix_sv_impl(
    self: CorrelationMatrix,
    *,
    config: EmulationConfig,
    state: StateVector,
    hamiltonian: DenseOperator,
) -> torch.Tensor:
    """
    Custom implementation of the density-density correlation ❬ψ|nᵢnⱼ|ψ❭
      for the state vector solver.
    TODO: extend to arbitrary two-point correlation ❬ψ|AᵢBⱼ|ψ❭
    """
    nqubits = state.n_qudits
    correlation = torch.zeros(
        nqubits, nqubits, dtype=torch.float64, device=state.vector.device
    )

    for i in range(nqubits):
        select_i = state.vector.view(2**i, 2, -1)
        select_i = select_i[:, 1]
        for j in range(i, nqubits):  # select the upper triangle
            if i == j:
                value = torch.linalg.vector_norm(select_i) ** 2
                correlation[j, j] = value
            else:
                select_i = select_i.view(2**i, 2 ** (j - i - 1), 2, -1)
                select_ij = select_i[:, :, 1, :]
                value = torch.linalg.vector_norm(select_ij) ** 2
                correlation[i, j] = value
                correlation[j, i] = value

    return correlation.cpu()


def energy_variance_sv_impl(
    self: EnergyVariance,
    *,
    config: EmulationConfig,
    state: StateVector,
    hamiltonian: RydbergHamiltonian,
) -> torch.Tensor:
    """
    Custom implementation of the energy variance ❬ψ|H²|ψ❭-❬ψ|H|ψ❭² for the state vector solver.
    """
    hstate = hamiltonian * state.vector
    h_squared = torch.vdot(hstate, hstate).real
    energy = torch.vdot(state.vector, hstate).real
    en_var: torch.Tensor = h_squared - energy**2
    return en_var.cpu()


def energy_second_moment_sv_impl(
    self: EnergySecondMoment,
    *,
    config: EmulationConfig,
    state: StateVector,
    hamiltonian: RydbergHamiltonian,
) -> torch.Tensor:
    """
    Custom implementation of the second moment of energy ❬ψ|H²|ψ❭
    for the state vector solver.
    """
    hstate = hamiltonian * state.vector
    en_2_mom: torch.Tensor = torch.vdot(hstate, hstate).real
    return en_2_mom.cpu()


def energy_sv_impl(
    self: Energy,
    *,
    config: EmulationConfig,
    state: StateVector,
    hamiltonian: RydbergHamiltonian,
) -> torch.Tensor:
    """
    Custom implementation of the energy ❬ψ|H|ψ❭ for the state vector solver.
    """
    en: torch.Tensor = hamiltonian.expect(state)
    return en



from __future__ import annotations
from collections import Counter
import math
from typing import Mapping, TypeVar, Type, Sequence
import torch
from pulser.backend import State
from emu_base import DEVICE_COUNT
from emu_sv.state_vector import StateVector
from emu_sv.utils import index_to_bitstring
from pulser.backend.state import Eigenstate

DensityMatrixType = TypeVar("DensityMatrixType", bound="DensityMatrix")

dtype = torch.complex128


class DensityMatrix(State[complex, torch.Tensor]):
    """Represents a density matrix in a computational basis."""

    # for the moment no need to check positivity and trace 1
    def __init__(
        self,
        matrix: torch.Tensor,
        *,
        gpu: bool = True,
    ):
        # NOTE: this accepts also zero matrices.

        device = "cuda" if gpu and DEVICE_COUNT > 0 else "cpu"
        self.matrix = matrix.to(dtype=dtype, device=device)

    @property
    def n_qudits(self) -> int:
        """The number of qudits in the state."""
        nqudits = math.log2(self.matrix.shape[0])
        return int(nqudits)

    @classmethod
    def make(cls, n_atoms: int, gpu: bool = True) -> DensityMatrix:
        """Creates the density matrix of the ground state |000...0>"""
        result = torch.zeros(2**n_atoms, 2**n_atoms, dtype=dtype)
        result[0, 0] = 1.0
        return cls(result, gpu=gpu)

    def __add__(self, other: State) -> DensityMatrix:
        raise NotImplementedError("Not implemented")

    def __rmul__(self, scalar: complex) -> DensityMatrix:
        raise NotImplementedError("Not implemented")

    def _normalize(self) -> None:
        # NOTE: use this in the callbacks
        """Normalize the density matrix state"""
        matrix_trace = torch.trace(self.matrix)
        if not torch.allclose(matrix_trace, torch.tensor(1.0, dtype=torch.float64)):
            self.matrix = self.matrix / matrix_trace

    def overlap(self, other: State) -> torch.Tensor:
        """
        Compute Tr(self^† @ other). The type of other must be DensityMatrix.

        Args:
            other: the other state

        Returns:
            the inner product

        Example:
        >>> density_bell_state = (1/2* torch.tensor([[1, 0, 0, 1], [0, 0, 0, 0],
        ... [0, 0, 0, 0], [1, 0, 0, 1]],dtype=torch.complex128))
        >>> density_c = DensityMatrix(density_bell_state, gpu=False)
        >>> density_c.overlap(density_c)
        tensor(1.+0.j, dtype=torch.complex128)
        """

        assert isinstance(
            other, DensityMatrix
        ), "Other state also needs to be a DensityMatrix"
        assert (
            self.matrix.shape == other.matrix.shape
        ), "States do not have the same number of sites"

        return torch.vdot(self.matrix.flatten(), other.matrix.flatten())

    @classmethod
    def from_state_vector(cls, state: StateVector) -> DensityMatrix:
        """Convert a state vector to a density matrix.
        This function takes a state vector |ψ❭ and returns the corresponding
        density matrix ρ = |ψ❭❬ψ| representing the pure state |ψ❭.
        Example:
           >>> from emu_sv import StateVector
           >>> import math
           >>> bell_state_vec = 1 / math.sqrt(2) * torch.tensor(
           ... [1.0, 0.0, 0.0, 1.0j],dtype=torch.complex128)
           >>> bell_state = StateVector(bell_state_vec, gpu=False)
           >>> density = DensityMatrix.from_state_vector(bell_state)
           >>> print(density.matrix)
           tensor([[0.5000+0.0000j, 0.0000+0.0000j, 0.0000+0.0000j, 0.0000-0.5000j],
                   [0.0000+0.0000j, 0.0000+0.0000j, 0.0000+0.0000j, 0.0000+0.0000j],
                   [0.0000+0.0000j, 0.0000+0.0000j, 0.0000+0.0000j, 0.0000+0.0000j],
                   [0.0000+0.5000j, 0.0000+0.0000j, 0.0000+0.0000j, 0.5000+0.0000j]],
                  dtype=torch.complex128)
        """

        return cls(
            torch.outer(state.vector, state.vector.conj()), gpu=state.vector.is_cuda
        )

    @classmethod
    def _from_state_amplitudes(
        cls: Type[DensityMatrixType],
        *,
        eigenstates: Sequence[Eigenstate],
        amplitudes: Mapping[str, complex],
    ) -> tuple[DensityMatrix, Mapping[str, complex]]:
        """Transforms a state given by a string into a density matrix.

        Construct a state from the pulser abstract representation
        https://pulser.readthedocs.io/en/stable/conventions.html

        Args:
            basis: A tuple containing the basis states (e.g., ('r', 'g')).
            nqubits: the number of qubits.
            strings: A dictionary mapping state strings to complex or floats amplitudes.

        Returns:
            The resulting state.

        Examples:
            >>> eigenstates = ("r","g")
            >>> n = 2
            >>> dense_mat=DensityMatrix.from_state_amplitudes(eigenstates=eigenstates,
            ... amplitudes={"rr":1.0,"gg":1.0})
            >>> print(dense_mat.matrix)
            tensor([[0.5000+0.j, 0.0000+0.j, 0.0000+0.j, 0.5000+0.j],
                    [0.0000+0.j, 0.0000+0.j, 0.0000+0.j, 0.0000+0.j],
                    [0.0000+0.j, 0.0000+0.j, 0.0000+0.j, 0.0000+0.j],
                    [0.5000+0.j, 0.0000+0.j, 0.0000+0.j, 0.5000+0.j]],
                   dtype=torch.complex128)
        """

        state_vector, amplitudes = StateVector._from_state_amplitudes(
            eigenstates=eigenstates, amplitudes=amplitudes
        )

        return DensityMatrix.from_state_vector(state_vector), amplitudes

    def sample(
        self,
        num_shots: int = 1000,
        one_state: Eigenstate | None = None,
        p_false_pos: float = 0.0,
        p_false_neg: float = 0.0,
    ) -> Counter[str]:
        """
        Samples bitstrings, taking into account the specified error rates.

        Args:
            num_shots: how many bitstrings to sample
            p_false_pos: the rate at which a 0 is read as a 1
            p_false_neg: teh rate at which a 1 is read as a 0

        Returns:
            the measured bitstrings, by count

        Example:
        >>> import math
        >>> torch.manual_seed(1234)
        >>> from emu_sv import StateVector
        >>> bell_vec = 1 / math.sqrt(2) * torch.tensor(
        ... [1.0, 0.0, 0.0, 1.0j],dtype=torch.complex128)
        >>> bell_state_vec = StateVector(bell_vec)
        >>> bell_density = DensityMatrix.from_state_vector(bell_state_vec)
        >>> bell_density.sample(1000)
         Counter({'00': 517, '11': 483})
        """

        assert p_false_neg == p_false_pos == 0.0, "Error rates must be 0.0"

        probabilities = torch.abs(self.matrix.diagonal())

        outcomes = torch.multinomial(probabilities, num_shots, replacement=True)

        # Convert outcomes to bitstrings and count occurrences
        counts = Counter(
            [index_to_bitstring(self.n_qudits, outcome) for outcome in outcomes]
        )

        return counts


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import torch

from emu_base.math.krylov_exp import krylov_exp
from emu_sv.hamiltonian import RydbergHamiltonian


def do_time_step(
    dt: float,
    omegas: torch.Tensor,
    deltas: torch.Tensor,
    phis: torch.Tensor,
    full_interaction_matrix: torch.Tensor,
    state_vector: torch.Tensor,
    krylov_tolerance: float,
) -> tuple[torch.Tensor, RydbergHamiltonian]:
    ham = RydbergHamiltonian(
        omegas=omegas,
        deltas=deltas,
        phis=phis,
        interaction_matrix=full_interaction_matrix,
        device=state_vector.device,
    )
    op = lambda x: -1j * dt * (ham * x)

    return (
        krylov_exp(
            op,
            state_vector,
            norm_tolerance=krylov_tolerance,
            exp_tolerance=krylov_tolerance,
        ),
        ham,
    )
